from typing import Dict, List

from dogebuild.relations import TaskRelationManager


class Context:
    def __init__(self, phases: Dict[str, List[str]] = None):
        self.relman = TaskRelationManager(phases)
        self.plugins = []
        self.dependencies = []
        self.test_dependencies = []


class ContextHolder:
    CONTEXT = None

    @staticmethod
    def create(phases=None):
        ContextHolder.CONTEXT = Context(phases)

    @staticmethod
    def clear_and_get():
        context = ContextHolder.CONTEXT
        ContextHolder.CONTEXT = None
        return context


def lifecycle(phases: Dict[str, List[str]] = None):
    ContextHolder.create(phases)



from typing import List
import os
import glob
from itertools import chain


DOGE_FILE = 'dogefile.py'


def files(base_dir: str, include: List[str], exclude: List[str] = None):
    if not exclude:
        exclude = []

    def expand_glob(base_dir, g):
        return glob.iglob(os.path.join(base_dir, g), recursive=True)

    return set(chain.from_iterable(map(lambda p: expand_glob(base_dir, p), include))) - set(chain.from_iterable(map(lambda p: expand_glob(base_dir, p), exclude)))



from typing import Dict, List, Callable, Tuple

from collections import OrderedDict
from toposort import toposort_flatten


class RelationManager:
    _DUMMY = 1

    def __init__(self):
        self._edges = {}

    def add_dependency(self, dependant, dependencies: List):
        if dependant not in self._edges:
            self._edges[dependant] = set()

        for dependency in dependencies:
            self._edges[dependant].add(dependency)

    def get_dependencies(self, key) -> List:
        return self._edges.get(key, [])

    def get_dependencies_recursive(self, keys: List) -> List:
        result = OrderedDict()

        def recursive_append(target: Dict, source: Dict, key):
            if key in source:
                target[key] = source[key]
                for x in target[key]:
                    if x not in target.keys():
                        recursive_append(target, source, x)

        for key in keys:
            dependencies = {}
            recursive_append(dependencies, self._edges, key)
            task_required = toposort_flatten(dependencies)
            for x in task_required:
                result[x] = self._DUMMY

        return list(result.keys())


class TaskRelationManager:
    DEFAULT_PHASES = {
        'clean': [],
        'docs': [],
        'docs-deploy': ['docs'],
        'sources': [],
        'resources': [],
        'test-sources': [],
        'test-resources': [],
        'test': ['sources', 'resources', 'test-sources', 'test-resources'],
        'build': ['test'],
        'run': ['build'],
        'integration-test': ['build'],
        'dist': ['integration-test'],
        'deploy': ['dist'],
        'install': ['dist'],
    }

    def __init__(self, phases: Dict[str, List[str]] = None):
        self._relation_manager = RelationManager()
        self._tasks = dict()

        if not phases:
            phases = TaskRelationManager.DEFAULT_PHASES

        for task, dependencies in phases.items():
            self.add_task(task, _skip)
            self.add_dependency(task, dependencies)

        self.verify()

    def add_task(self, task_name: str, task: Callable):
        self._tasks[task_name] = task
        self._relation_manager.add_dependency(task_name, [])

    def add_dependency(self, dependant: str, dependencies: List[str]):
        self._relation_manager.add_dependency(dependant, dependencies)

    def get_dependencies(self, dependant: str) -> List[str]:
        return self._relation_manager.get_dependencies(dependant)

    def get_tasks(self, task_names: List[str]) -> List[Tuple[str, Callable]]:
        task_names = self._relation_manager.get_dependencies_recursive(task_names)
        return list(map(lambda name: (name, self._tasks[name]), task_names))

    def verify(self):
        known_task_names = set()
        for key in self._relation_manager._edges.keys():
            known_task_names.add(key)
            for dep in self._relation_manager._edges[key]:
                known_task_names.add(dep)

        for task_name in known_task_names:
            if task_name not in self._tasks.keys():
                raise Exception("Inconsistent task graph: unknown name '{}'".format(task_name))


def _skip():
    return 0, {}



import logging
from os import path
from typing import List, Optional

from .common import DOGE_FILE
from .dependencies import Dependency
from .dogefile_loader import load_doge_file


def resolve_dependency_tree(dependencies: List[Dependency], parents: List[str] = None) -> List[Dependency]:
    if not parents:
        parents = []

    for d in dependencies:
        id, version = d.get_id()
        if id in parents:
            raise Exception('Circular dependency')
        use_version = _resolve_version_(id, version)
        if version != use_version:
            d.original_version = d.version
            d.version = use_version
        logging.info('Acquiring {} ...'.format(d))
        d.acquire_dependency()
        deps, _ = load_doge_file(path.join(d.get_doge_file_folder(), DOGE_FILE))
        d.dependencies = resolve_dependency_tree(deps, parents + [id])
    return dependencies


VERSIONS = {}


def _resolve_version_(id: str, version: Optional[str]) -> Optional[str]:
    if not version:
        return None

    # simple conflict resolving strategy
    saved_version = VERSIONS.get(id)
    if not saved_version:
        VERSIONS[id] = version
        return version
    else:
        return saved_version



from subprocess import check_call
import os
from os import path
from shutil import rmtree
from typing import Tuple, Optional

from dogebuild.context import ContextHolder


class Dependency:
    def __init__(self, **kwargs):
        self.context = kwargs

    def get_id(self) -> Tuple[str, Optional[str]]:
        raise NotImplementedError

    def __str__(self):
        id, version = self.get_id()
        if not version:
            return id
        else:
            return '{} ({})'.format(id, version)

    def acquire_dependency(self):
        raise NotImplementedError()


class GitDependency(Dependency):
    GIT_REPO_FOLDER = path.expanduser(path.join('~', '.doge', 'repo', 'git'))

    VERSION_TAG = 'tag:'
    VERSION_BRANCH = 'branch:'
    # TODO: make VERSION_COMMIT

    def __init__(self, url, version, **kwargs):
        super().__init__(**kwargs)

        self.url = url
        self.version = version
        self.original_version = None
        self.name = url.split('/')[-1].replace('.git', '')
        self.org = url.split('/')[-2].split(':')[-1]
        self.dependencies = []

    def acquire_dependency(self):
        self._assert_repo_folder_exists()

        if self.version.startswith(GitDependency.VERSION_TAG):
            tag = self.version[len(GitDependency.VERSION_TAG):]
            tag_folder = path.join(GitDependency.GIT_REPO_FOLDER, self.org, self.name, 'tags', tag)
            if not path.exists(tag_folder):
                os.makedirs(tag_folder)
                check_call(['git', 'clone', '-b', tag, self.url, '.'], cwd=tag_folder)
                rmtree(path.join(tag_folder, '.git'))

        elif self.version.startswith(GitDependency.VERSION_BRANCH):
            branch = self.version[len(GitDependency.VERSION_BRANCH):]
            branch_folder = path.join(GitDependency.GIT_REPO_FOLDER, self.org, self.name, 'branches', branch)
            if not path.exists(branch_folder):
                os.makedirs(branch_folder)
                check_call(['git', 'clone', '-b', branch, self.url, '.'], cwd=branch_folder)
            else:
                check_call(['git', 'checkout', branch], cwd=branch_folder)
                check_call(['git', 'pull', '--ff-only'], cwd=branch_folder)

        else:
            raise NotImplementedError

    def get_doge_file_folder(self):
        if self.version.startswith(GitDependency.VERSION_TAG):
            tag = self.version[len(GitDependency.VERSION_TAG):]
            return path.join(GitDependency.GIT_REPO_FOLDER, self.org, self.name, 'tags', tag)

        elif self.version.startswith(GitDependency.VERSION_BRANCH):
            branch = self.version[len(GitDependency.VERSION_BRANCH):]
            return path.join(GitDependency.GIT_REPO_FOLDER, self.org, self.name, 'branches', branch)

        else:
            raise NotImplementedError

    def get_id(self) -> Tuple[str, Optional[str]]:
        return self.url, self.version

    def _assert_repo_folder_exists(self):
        os.makedirs(GitDependency.GIT_REPO_FOLDER, exist_ok=True)


class FolderDependency(Dependency):
    def __init__(self, folder, **kwargs):
        super().__init__(**kwargs)

        self.folder = folder
        self.version = None
        self.original_version = None
        self.dependencies = []

    def acquire_dependency(self):
        pass

    def get_doge_file_folder(self):
        return self.folder

    def get_id(self) -> Tuple[str, Optional[str]]:
        return self.folder, None


def folder(folder: str, **kwargs) -> FolderDependency:
    return FolderDependency(folder, **kwargs)


def git(repo: str, version: str='branch:master', **kwargs) -> GitDependency:
    return GitDependency(repo, version, **kwargs)


def dependencies(*args):
    ContextHolder.CONTEXT.dependencies += args


def test_dependencies(*args):
    ContextHolder.CONTEXT.test_dependencies += args



from dogebuild.plugins import ContextHolder
from dogebuild.common import DOGE_FILE


def load_doge_file(filename):
    with open(filename) as f:
        code = compile(f.read(), DOGE_FILE, 'exec')

        ContextHolder.create()
        exec(code)
        return ContextHolder.clear_and_get()



from typing import List

from dogebuild.context import ContextHolder


def task(name: str = None, depends: List[str] = None, phase: str = None):
    if not depends:
        depends = []

    relman = ContextHolder.CONTEXT.relman

    def decorator(task_callable):
        task_name = name
        if not task_name:
            task_name = task_callable.__name__
        task_name = 'dogefile:' + _sanitize_name(task_name)

        relman.add_task(task_name, task_callable)
        relman.add_dependency(task_name, depends)

        if phase:
            relman.add_dependency(phase, [task_name])

        return task_callable

    return decorator


def _sanitize_name(name: str):
    return name.replace('_', '-')



from logging import getLogger
from typing import Callable, List

from dogebuild.context import ContextHolder


class DogePlugin:
    NAME = 'This is abstract doge plugin so this variable should never be used'

    @classmethod
    def get_name(cls):
        return cls.NAME

    def __init__(self, **kwargs):
        self.logger = getLogger(self.get_name())

        ContextHolder.CONTEXT.plugins.append(self)

        self.relman = ContextHolder.CONTEXT.relman
        self.dependencies = ContextHolder.CONTEXT.dependencies
        self.test_dependencies = ContextHolder.CONTEXT.test_dependencies

    def add_task(self, task_name: str, task: Callable, phase: str = None):
        task_name = self._resolve_full_task_name(task_name)

        self.relman.add_task(task_name, task)
        if phase:
            self.relman.add_dependency(phase, [task_name])
            self.relman.add_dependency(task_name, self.relman.get_dependencies(phase))

    def add_dependency(self, task_name: str, dependencies: List[str]):
        task_name = self._resolve_full_task_name(task_name)
        dependencies = list(map(lambda n: self._resolve_full_task_name(n), dependencies))

        self.relman.add_dependency(task_name, dependencies)

    def _resolve_full_task_name(self, task_name: str):
        return self.__class__.get_name() + ":" + task_name



import argparse
import sys
import os
from typing import List, Callable, Optional, Tuple, Dict
import logging
import logging.config

from dogebuild.common import DOGE_FILE
from dogebuild.dependencies import Dependency
from dogebuild.dependencies_functions import resolve_dependency_tree
from dogebuild.dogefile_loader import load_doge_file


def _main() -> None:
    _config_logging()

    main_parser = argparse.ArgumentParser(description='')
    main_parser.add_argument('command', nargs=1)
    main_parser.add_argument('options', nargs=argparse.REMAINDER)

    main_args = main_parser.parse_args()

    func = _get_function(main_args.command[0])

    if callable(func):
        exit(func(*main_args.options))
    else:
        tasks = main_args.command + main_args.options
        tasks = map(lambda name: name.lstrip(':'), tasks)
        exit(_run_tasks(*tasks))


def _get_function(name: str) -> Optional[Callable]:
    name = name.lower().strip().replace('-', '_')
    if not name[0].isalpha():
        return None
    else:
        return getattr(_get_current_module(), name, None)


def _get_current_module():
    module_name = globals()['__name__']
    return sys.modules[module_name]


def dependency_tree() -> int:
    deps, _ = load_doge_file(DOGE_FILE)
    deps = resolve_dependency_tree(deps)
    _print_dependencies(deps)
    return 0


def _print_dependencies(dependencies: List[Dependency], inner_level: int=0):
    for d in dependencies:
        if d.original_version:
            print(' ' * (2 * inner_level - 1) + '+' + str(d) + ' conflict resolved for {}'.format(d.original_version))
        else:
            print(' ' * (2 * inner_level - 1) + '+' + str(d))
        _print_dependencies(d.dependencies, inner_level=inner_level + 1)


def _run_tasks(*tasks) -> int:
    return _run_task_of_file(DOGE_FILE, *tasks)[0]


def _run_task_of_file(doge_file, *tasks) -> Tuple[int, Dict]:
    abs_path = os.path.abspath(doge_file)
    doge_directory = os.path.dirname(abs_path)
    doge_file_name = os.path.basename(abs_path)

    context = load_doge_file(abs_path)

    relman = context.relman
    relman.verify()

    dependencies = context.dependencies
    test_dependencies = context.test_dependencies

    for dependency in dependencies + test_dependencies:
        logging.info('Resolving dependency {}'.format(dependency))
        dependency.acquire_dependency()
        code, artifacts = _run_task_of_file(os.path.join(dependency.get_doge_file_folder(), DOGE_FILE), 'build')
        if not code:
            dependency.artifacts = artifacts
        else:
            logging.error('Dependency {} build failed'.format(dependency))
            return code, {}

    run_list = relman.get_tasks(tasks)
    logging.info('Run tasks: {}'.format(', '.join(map(lambda x: x[0], run_list))))

    os.chdir(doge_directory)
    artifacts = {}
    for t in run_list:
        exit_code, current_artifacts = t[1]()
        if not exit_code:
            logging.debug('Task {} successfully terminated'.format(t[0]))
            _add_artifacts(artifacts, current_artifacts)
        else:
            logging.error('Task {} failed'.format(t[0]))
            return exit_code, {}

    return 0, artifacts


def _add_artifacts(main: Dict, add: Dict):
    for type, artifacts in add.items():
        if type in main:
            main[type] += artifacts
        else:
            main[type] = artifacts


def _config_logging():
    console_format = '{log_color}{name}: {message}{reset}'
    console_level = 'DEBUG'

    logging.config.dictConfig({
        'version': 1,
        'formatters': {
            'colored': {
                '()': 'colorlog.ColoredFormatter',
                'format': console_format,
                'style': '{',
            },
        },
        'handlers': {
            'console': {
                'class': 'logging.StreamHandler',
                'formatter': 'colored',
                'level': console_level,
            },
        },
        'loggers': {
            '': {
                'handlers': ['console'],
                'level': console_level,
            },
        },
    })



from setuptools import setup, find_packages
from os import path


project_directory = path.abspath(path.dirname(__file__))


def load_from(file_name):
    with open(path.join(project_directory, file_name), encoding='utf-8') as f:
        return f.read()


setup(
    name='dogebuild',
    version=load_from('dogebuild/dogebuild.version').strip(),
    description='Builder with plugin system',
    long_description=load_from('README.md'),
    long_description_content_type='text/markdown',
    url='https://github.com/dogebuild/dogebuild',
    author='Kirill Sulim',
    author_email='kirillsulim@gmail.com',
    license='MIT',
    packages=find_packages(include=[
        'dogebuild',
    ]),
    package_data={
        'dogebuild': [
            'dogebuild.version',
        ]
    },
    scripts=[
        'scripts/doge_runner.py',
    ],
    entry_points={
        'console_scripts': [
            'doge = doge_runner:main',
        ],
    },
    test_suite='tests',
    install_requires=[
        'pip',
        'virtualenv',
        'toposort==1.5',
        'colorlog==4.0.2',
    ],
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Environment :: Console',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Natural Language :: English',
        'Operating System :: OS Independent',
        'Topic :: Software Development',
    ],
    keywords='dogebuild builder',
)
